#include "nios.h"
#include "stdio.h"
#include <time.h>
#include <sys/time.h>

/*********************************************/
/* home-made clock routines ...              */
/*********************************************/
long nr_clockOffset = 0;

long timezone;   /* normally in time.h */

void tzset(void) /* normally in time.h */
{
  /* fill this in some day ...*/
}
/* These two functions below are important ! */
long time(long *lptime)
{
  struct timeval tv;
  gettimeofday(&tv,(struct timezone *)NULL);
  if (lptime != NULL) *lptime = tv.tv_sec;
  return tv.tv_sec;
}
/* running count since last boot */
int _gettimeofday(struct timeval *tv,struct timezone *tz)
{
  static int tsec = 0;  /* adjusted offset for potential wrapping of tmsec */
  static int tmsec = 0; /* result of most recent call */
  int nt = nr_timer_milliseconds(); /* this call */
  if (nt < tmsec) 
  {
    tsec += tmsec/1000; /* it has wrapped */
    nr_clockOffset += tsec;
  }
  tmsec = nt;
  if (tv != NULL)
  {
    tv->tv_sec = nr_clockOffset + tmsec/1000;
    tv->tv_usec = (tmsec % 1000) * 1000;
  }
  return 0;
}
/* function to 'set the clock' ... */
int nr_setclock(long tgtTime)
{
  long dt, tnow = time(NULL);
  dt = tgtTime - tnow;
  nr_clockOffset += dt;
  if (dt > 10 || dt < -10)
  {
    printf("nr_setclock: major time jump %d to %d\n",tnow,tgtTime);
  }
  return dt;
}

/*********************************************/
/* cover routines for file IO ...            */
/*********************************************/
int _open(const char *filename,int oflag,int pmode)
{
  return -1; /* return failure */
}
int _unlink(const char *filename)
{
  return 0;
}
int _link(const char *filename)
{
  return -1;
}
int nr_fprintf(FILE *fp, const char *fmt, ... )
{
	return 0; /* return failure */
}

/*********************************************/
/* memory management routines ...            */
/*********************************************/
void *SystemCalloc(size_t nr, size_t size)
{
  size_t fullsize = nr*size;
  void *ptr;
  ptr = malloc(fullsize);
  if (ptr == NULL) printf("malloc null %x (size=%d)\n", ptr, fullsize);
  if (ptr != NULL) memset( ptr, 0, fullsize );
  return( ptr );
}
void *SystemMalloc( size_t fullsize )
{
  void *ptr;
  ptr = malloc( fullsize );
  if (ptr == NULL) printf("malloc null %x (size=%d)\n", ptr, fullsize);
  return( ptr );
}
void SystemFree( void *ptr )
{
  if (ptr == NULL) printf("free null %x\n", ptr);
  free( ptr );
}
