#include "project.def"
#include "includes.h"
#include "devices.h"
#include <string.h>
#include "../inc/lwip/tcpip.h"
#include "../inc/lwip/inet.h"
#include "../inc/lwip/tcp.h"
#include "../inc/lwip/udp.h"
#include <fcntl.h>
#include "../inc/lwip/mem.h"

int debuglevel = 1;
int lwip_debuglevel = 0; // debug level of lwip
SckMemPool sckPool[MAX_NUM_SCK];
int nSckInPool = 0;
int lastSocketError = 0;
#define ccerr(i) { cc=i; goto err; }
//#define TRUE -1
//#define FALSE 0
#define MAX(a,b) ((a)>(b)?a:b)
#define MIN(a,b) ((a)<(b)?a:b)

void doLWIPInit();


void sleep(int msec) 
{ 
  INT8U  lsec =  (INT8U)  (msec / 1000) % 60;
  INT16U lmsec = (INT16U) (msec % 1000);
  INT8U  lmin =  (INT8U)  (((msec / 1000) / 60) % 60);
  INT8U  lhour = (INT8U)  (((msec / 1000) / 60) / 60);

  OSTimeDlyHMSM(lhour,lmin,lsec,lmsec);

  // SW! nr_delay(msec); 
}

char *inet_ntoa(struct in_addr in)
{
  static char ipstr[32];

  sprintf(ipstr,"%d.%d.%d.%d",
  (u_char)in.S_un.S_un_b.s_b1,
  (u_char)in.S_un.S_un_b.s_b2,
  (u_char)in.S_un.S_un_b.s_b3,
  (u_char)in.S_un.S_un_b.s_b4);

  return ipstr;
}

// SW! added function
unsigned long makeIP(int a, int b, int c, int d)
{
 return( ((a&0xFF)<<24) + ((b&0xFF)<<16) +((c&0xFF)<<8) + (d&0xFF) );
}

// SW! returns NULL now since lwIP does not support DNS lookup, only IP adresses are supported
struct hostent *gethostbyname(const char *hstname)
{
  /* SW! static struct hostent hst;

  hst.h_name = (char *)hstname;
  hst.h_aliases = 0;
  hst.h_addrtype = AF_INET;
  hst.h_length = 4;
  hst.h_addr = 0;

  return &hst;
  */
  return NULL; // SW!
}

struct hostent *gethostbyaddr(const char *hstaddr, int hstlen, int hsttype)
{
  static struct hostent hst;
  static char haddr[4];
  char *c;
  int i = 0;

  hst.h_name = (char *)hstaddr;
  hst.h_aliases = 0;
  hst.h_addrtype = hsttype;
  hst.h_length = hstlen;

  for (c=strtok((char *)hstaddr,"."); c != NULL; c=strtok((char *)NULL,".")) // "string.h" needed? grev
  {
    haddr[3-i++] = atoi(c);
  }

  hst.h_addr = haddr;

  return &hst;
}
unsigned long inet_addr (char *cp)
{
  struct hostent *hst = gethostbyaddr(cp,strlen(cp),AF_INET);
  return hst == NULL ? 0 : (unsigned long) hst->h_addr;
}
/*
sck [in] Descriptor identifying a connected socket. 
name [out] The structure that receives the name of the peer. 
namelen [in, out] Pointer to the size of the name structure. 
*/
int getpeername(SOCKET sck,struct sockaddr *name,int *namelen)
{
  int i, cc=0, sckPtr;

  if (sck < 0 || sck >= nSckInPool) ccerr(ENOTSOCK);
  if (namelen == NULL || ((unsigned int) *namelen) < sizeof(struct sockaddr)) ccerr(EINVAL);
  sckPtr = sckPool[sck].sndBufferHead;
  memcpy(name,&sckPool[sck].sckRcvPackets[sckPtr].sckAddr,sizeof(struct sockaddr));
err:
  lastSocketError = cc;
  if (debuglevel) printf("connect : return <%d>\n",cc);
  return cc ? -1 : 0;
}
/* commented out, provided by lwip stack
// SW! translated to work everywhere
unsigned short htons (unsigned short n) 
{ 
  return( ((n & 0xFF)<< 8) + ((n & 0xFF00)>>8) );
  //return nr_h2n16(n); 
}

unsigned long htonl (unsigned long n) 
{ 
  return( ((n & 0xFF000000)>>24) + (( n & 0xFF0000)>>8) + ((n & 0xFF00)<< 8) + ((n & 0xFF)<<24) );
  //return nr_h2n32(n); 
}

unsigned short ntohs (unsigned short n) 
{ 
  //return( ((n & 0xFF)<< 8) + ((n & 0xFF00)>>8) );
  return( htons(n) );
  //return nr_n2h16(n); 
} 

unsigned long  ntohl (unsigned long n) 
{ 
  //return( ((n & 0xFF000000)>>24) + (( n & 0xFF0000)>>8) + ((n & 0xFF00)<< 8) + ((n & 0xFF)<<24) );
  return( htonl( n ) );
  // return nr_n2h32(n); 
}
*/
int socket(int af, int type, int protocol)
{
 int cc = 0, sck = -1;
 if (nSckInPool >= MAX_NUM_SCK) ccerr(ENOBUFS);
 sck = nSckInPool++;
 sckPool[sck].typ = type;
 sckPool[sck].flg = lwip_debuglevel; // SW!
 sckPool[sck].sckRcvPackets = NULL; // SW! 19.11.03
 sckPool[sck].sckSndPackets = NULL; // SW! 19.11.03

 /* sckPool[sck].sck will get filled in with the nios plug at socketInit() time */

err:
 if (debuglevel) printf("socket : return %d <%d>\n",sck,cc);
 lastSocketError = cc;
 return sck;
}

// SW! added function

void lwipUDPCallback( void *arg, struct udp_pcb *upcb, struct pbuf *p, struct ip_addr *addr, u16_t port)
{
  int i, cc=0;
  udp_status *status = arg;				
  int sckPtr;
  char err;
  
  void *payload = p->payload; // SW! 17.11.03
  int payload_length = p->len; // SW! 17.11.03

  //printf("UDP callback called\n"); // SW! 19.11.03

  if (payload_length > RCVBUF_MTU) ccerr(EMSGSIZE);
  for (i=0; i<nSckInPool && upcb != sckPool[i].udppcb; i++);
  if (i == nSckInPool) ccerr(ENOTSOCK);
  if (sckPool[i].nrcvready == sckPool[i].numRcvBuffers) ccerr(ENOBUFS);
  sckPtr = sckPool[i].rcvBufferHead;

  sckPool[i].nrcvready++;
  sckPool[i].sckRcvPackets[sckPtr].sckAddr.sin_family = PF_INET; /* hardwire this */
  sckPool[i].sckRcvPackets[sckPtr].h_addr = addr->addr;
  sckPool[i].sckRcvPackets[sckPtr].sckAddr.sin_port = port;
  sckPool[i].sckRcvPackets[sckPtr].sckAddr.sin_addr.S_un.S_un_b.s_b1 = (addr->addr)&0xFF;
  sckPool[i].sckRcvPackets[sckPtr].sckAddr.sin_addr.S_un.S_un_b.s_b2 = (addr->addr >> 8)&0xFF;
  sckPool[i].sckRcvPackets[sckPtr].sckAddr.sin_addr.S_un.S_un_b.s_b3 = (addr->addr >> 16)&0xFF;
  sckPool[i].sckRcvPackets[sckPtr].sckAddr.sin_addr.S_un.S_un_b.s_b4 = (addr->addr >> 24)&0xFF;
  memset(sckPool[i].sckRcvPackets[sckPtr].sckAddr.sin_zero,0,8);

  memcpy(sckPool[i].sckRcvPackets[sckPtr].buf,payload,payload_length);
  sckPool[i].sckRcvPackets[sckPtr].bufMsgSize = payload_length;

  err = 1;
  while (err > 0)
  {
   err = pbuf_free( p );
   printf("pbuf_free UDPcallb %d\n", err);
  }
 
  if (debuglevel) printf("sck %d ptr %d udpCallback from %d.%d.%d.%d:%d %d bytes\n",
	  i,sckPtr,addr->addr&0xFF,(addr->addr >> 8)&0xFF,(addr->addr >> 16)&0xFF,(addr->addr >> 24)&0xFF,port,
          sckPool[i].sckRcvPackets[sckPtr].bufMsgSize);

  sckPool[i].rcvBufferHead = (sckPool[i].rcvBufferHead + 1) % sckPool[i].numRcvBuffers;
err:
  lastSocketError = cc;
  if (debuglevel) printf("plugsCallback : return <%d>\n",cc);

  // return cc ? -1 : 0; // SW! removed 
}

err_t lwipTCPCallback( void *arg, struct tcp_pcb *tpcb, struct pbuf *p, err_t err)
{

}


/* cache the incoming data, sender address, raise a flag for select => move head (head-tail -> flag) */
/* SW! removed due to no plugs any more
int plugsCallback(int plug_handle, void *context, ns_plugs_packet *p, void *payload, int payload_length)
{
  int i, cc=0;
  void *udpheader = p[ne_plugs_udp].header;
  void *ipheader = p[ne_plugs_ip].header;
  unsigned long ip = ((unsigned int *)ipheader)[3];
  unsigned short port = ((unsigned short *)udpheader)[0];
  u_char *ipaddrblk = (u_char *)ipheader;
  udp_status *status = context;				
  net_32 iptosendto;
  net_16 porttosendto;
  int sckPtr;

  if (payload_length > RCVBUF_MTU) ccerr(EMSGSIZE);
  for (i=0; i<nSckInPool && plug_handle != sckPool[i].sck; i++);
  if (i == nSckInPool) ccerr(ENOTSOCK);
  if (sckPool[i].nrcvready == sckPool[i].numRcvBuffers) ccerr(ENOBUFS);
  sckPtr = sckPool[i].rcvBufferHead;
  sckPool[i].nrcvready++;
  sckPool[i].sckRcvPackets[sckPtr].sckAddr.sin_family = PF_INET; // hardwire this
  sckPool[i].sckRcvPackets[sckPtr].h_addr = ntohl(ip);
  sckPool[i].sckRcvPackets[sckPtr].sckAddr.sin_port = nr_n2h16(port);
  sckPool[i].sckRcvPackets[sckPtr].sckAddr.sin_addr.S_un.S_un_b.s_b1 = ipaddrblk[12];
  sckPool[i].sckRcvPackets[sckPtr].sckAddr.sin_addr.S_un.S_un_b.s_b2 = ipaddrblk[13];
  sckPool[i].sckRcvPackets[sckPtr].sckAddr.sin_addr.S_un.S_un_b.s_b3 = ipaddrblk[14];
  sckPool[i].sckRcvPackets[sckPtr].sckAddr.sin_addr.S_un.S_un_b.s_b4 = ipaddrblk[15];
  memset(sckPool[i].sckRcvPackets[sckPtr].sckAddr.sin_zero,0,8);

  memcpy(sckPool[i].sckRcvPackets[sckPtr].buf,payload,payload_length);
  sckPool[i].sckRcvPackets[sckPtr].bufMsgSize = payload_length;

  if (debuglevel) printf("sck %d ptr %d plugsCallback from %d.%d.%d.%d:%d %d bytes\n",
	  i,sckPtr,ipaddrblk[12],ipaddrblk[13],ipaddrblk[14],ipaddrblk[15],port,sckPool[i].sckRcvPackets[sckPtr].bufMsgSize);

  sckPool[i].rcvBufferHead = (sckPool[i].rcvBufferHead + 1) % sckPool[i].numRcvBuffers;
err:
  lastSocketError = cc;
  if (debuglevel) printf("plugsCallback : return <%d>\n",cc);
  return cc ? -1 : 0;
}
*/

int socketInit(int sck)
{
  static int done=0;
  int type;
  int cc=0;
  struct sockaddr_in *sa;

  if (!done)
  {
   /* SW! commented out, not needed using lwip stack
	  if ((cc=nr_plugs_initialize(
		      0, // -> ???? ne_plugs_flag_dhcp, // important if board uses DHCP !!!
		      0,
		      __adapter__,
		      __adapter_irq__,
		      __adapter_struct_addr__
       )) != 0) ccerr(cc);
   */
    doLWIPInit();
    done = TRUE;
    //if (debuglevel) printf("nr_plugs_initialized\n"); // SW! commented out 
    if (debuglevel) printf("lwIP stack initialized\n"); // SW! commented out 
  }

  if (sckPool[sck].initialized) return 0;
  switch (sckPool[sck].typ)
  {
    case SOCK_DGRAM:
      sckPool[sck].udppcb = udp_new();
      if (sckPool[sck].udppcb == NULL) ccerr(ENOMEM);
      //printf("UDP Receive\n"); 
      udp_recv(sckPool[sck].udppcb, lwipUDPCallback, &sckPool[sck].sts);
      //type = ne_plugs_udp;
      break;
    case SOCK_STREAM:
      sckPool[sck].tcppcb = tcp_new();
      if (sckPool[sck].tcppcb == NULL) ccerr(ENOMEM);
      tcp_recv(sckPool[sck].tcppcb, lwipTCPCallback);
      tcp_arg(sckPool[sck].tcppcb, &sckPool[sck].sts);
      //type = ne_plugs_tcp;
      break;
    default:
      ccerr(ENOPROTOOPT);
      break;
  }

/* SW! commented out 17.11.03
  if ((cc=nr_plugs_create(
		    &sckPool[sck].sck,
		    type,
		    sckPool[sck].name.sin_port,
		    plugsCallback,
		    &sckPool[sck].sts,
		    sckPool[sck].flg
		  )) != 0) ccerr(cc);
*/

/* SW! commented out 17.11.03
  if (debuglevel) 
  {
	printf("nr_plugs_create\n");
	printf("sck   : %d\n",sck);
	printf("type  : %d\n",type);
	printf("port  : %d\n",sckPool[sck].name.sin_port);
	printf("plug  : %d\n",sckPool[sck].sck);
	printf("status: %d\n",sckPool[sck].sts);
  }
  if ((cc=nr_plugs_connect(
        sckPool[sck].sck,
        0,
        -1, 
        -1
      )) != 0) ccerr(cc);
*/

  sa = (struct sockaddr_in *)&sckPool[sck].name;
  sa->sin_addr.s_addr = makeIP(IPADDR0, IPADDR1, IPADDR2, IPADDR3);
  memset(sa->sin_zero,0,8);

  sckPool[sck].initialized = TRUE;
  if (debuglevel) printf("done sockinit\n"); // SW! changed 17.11.03

err:
  lastSocketError = cc;
  if (debuglevel) printf("socketInit : return <%d>\n",cc);
  return cc ? -1 : 0;
}

unsigned short nextFreePort(void)
{
  static unsigned short p = USER_PORT_BASE;
  int i, n;

  for (n=0; n<65535; n++)
  {
    for (i=0; i < nSckInPool && sckPool[i].name.sin_port != p; i++);
    if (i == nSckInPool) return p;
    p++;
  }
  return FALSE;		// 28.10.03, grev
}

int bind(SOCKET s,const struct sockaddr *name, int namelen)
{
  int i, cc=0;
  int optval;
  err_t binderr;
  struct sockaddr_in *sa = (struct sockaddr_in *)name;
  if (s < 0 || s >= nSckInPool) ccerr(ENOTSOCK);

  if (sa == NULL) 
  {
    sckPool[s].name.sin_family  = AF_INET;
    sckPool[s].name.sin_port = nextFreePort();
  }
  else
  {
    /* let's ignore the namelen for now ... */
    if (!(sckPool[s].opt & SO_REUSEADDR))
    {
      for (i=0; i < nSckInPool && sckPool[i].name.sin_port != sa->sin_port; i++);
      if (i < nSckInPool) ccerr(EADDRINUSE);
    }

    sckPool[s].name.sin_addr.s_addr = sa->sin_addr.s_addr;
    sckPool[s].name.sin_family  = sa->sin_family;
    sckPool[s].name.sin_port = ntohs(sa->sin_port);
  }

  /* the port is established, now make sure it's initialized */
  if (!sckPool[s].initialized && (cc=socketInit(s)) != 0) ccerr(cc);

  // bind the port 
  binderr = udp_bind(sckPool[s].udppcb, (struct ip_addr *) &(sa->sin_addr.s_addr), ntohs(sa->sin_port));

  if (binderr != ERR_OK) ccerr(binderr);

  if (sckPool[s].numRcvBuffers == 0) 
  {
    optval = DEFAULT_RCVBUF_SIZE;
    cc = setsockopt(s,SOL_SOCKET,SO_RCVBUF,(const char *)&optval,sizeof(int));
  }
  if (sckPool[s].numSndBuffers == 0) 
  {
    optval = DEFAULT_SNDBUF_SIZE;
    cc = setsockopt(s,SOL_SOCKET,SO_SNDBUF,(const char *)&optval,sizeof(int));
  }

err:
 lastSocketError = cc;
 if (debuglevel) printf("bind : socket %d return <%d>\n",s,cc);
 return cc ? -1 : 0;
}
  /*
  s 
  [in] Descriptor identifying a socket. 
  level 
  [in] Level at which the option is defined; the supported levels include SOL_SOCKET and IPPROTO_TCP. See the Windows Sockets 2 Protocol-Specific Annex (a separate document included with the Platform SDK) for more information on protocol-specific levels. 
  optname 
  [in] Socket option for which the value is to be set. 
  optval 
  [in] Pointer to the buffer in which the value for the requested option is supplied. 
  optlen 
  [in] Size of the optval buffer. 
  */
int setsockopt(SOCKET s, int level, int optname, const char *optval, int optlen)
{
  int i, n, cc=0;
  if (s < 0 || s >= nSckInPool) ccerr(ENOTSOCK);
  if (level != SOL_SOCKET) ccerr(EOPNOTSUPP);
  switch (optname)
  {
    case SO_RCVBUF:
 //     if (optlen >= sizeof(int)) n = *(int *)optval; else ccerr(EINVAL);
      if ((unsigned)optlen >= sizeof(int)) n = *(int *)optval; else ccerr(EINVAL);	//271003, grev
      if (sckPool[s].sckRcvPackets != NULL) ccerr(EALREADY);
      if ((sckPool[s].sckRcvPackets=(SckPacket *)SystemCalloc(n,sizeof(SckPacket))) == NULL) ccerr(ENOMEM);
      sckPool[s].numRcvBuffers = n;
      break;
    case SO_SNDBUF:
//      if (optlen >= sizeof(int)) n = *(int *)optval; else ccerr(EINVAL);
      if ((unsigned)optlen >= sizeof(int)) n = *(int *)optval; else ccerr(EINVAL);	//271003, grev
      if (sckPool[s].sckSndPackets != NULL) ccerr(EALREADY);
      if ((sckPool[s].sckSndPackets=(SckPacket *)SystemCalloc(n,sizeof(SckPacket))) == NULL) ccerr(ENOMEM);
      sckPool[s].numSndBuffers = n;
      break;
    default:
  }
err:
  lastSocketError = cc;
  if (debuglevel) printf("setsockopt : socket %d return <%d>\n",s,cc);
  return cc ? -1 : 0;
}

/*
Parameters
nfds [in] Ignored. The nfds parameter is included only for compatibility with Berkeley sockets. 
readfds [in/out] Optional pointer to a set of sockets to be checked for readability. 
writefds [in/out] Optional pointer to a set of sockets to be checked for writability 
exceptfds [in/out] Optional pointer to a set of sockets to be checked for errors. 
timeout [in] Maximum time for select to wait, provided in the form of a TIMEVAL structure. Set the timeout parameter to NULL for blocking operation. 
*/
int select(int nfds, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, const struct timeval *timeout)
{
  int i, nrd=0, nwrt=0, cc=0;
  fd_set rd, wr;

  for (i=0; i<nSckInPool; i++)
  {
     if (readfds != NULL)
     {
       rd = *readfds;
       FD_ZERO(readfds);
       if (FD_ISSET(i,&rd) && sckPool[i].nrcvready > 0) 
       {
         FD_SET(i,readfds);
         nrd++;
       }
     }
     if (writefds != NULL) 
     {
       wr = *writefds;
       FD_ZERO(writefds);
	   sckPool[i].nsndready = 1; // hard-wire this for now ...
       if (FD_ISSET(i,&wr) && sckPool[i].nsndready > 0) 
       {
         FD_SET(i,writefds);
         nwrt++;
       }
     }
  }
  
err:
  lastSocketError = cc;
  if (debuglevel > 1) printf("select : return <%d>\n",cc);
  return cc ? -1 : nrd + nwrt;
}
/*
Parameters
s [in] Descriptor identifying a socket. 
cmd [in] Command to perform on the socket s. 
argp [in/out] Pointer to a parameter for cmd. 
*/
int soioctl(SOCKET sck,long cmd, u_long *argp)
{
  int cc=0;
  /* used mostly to set NON-BLOCKING IO */
  
  if (sck < 0 || sck >= nSckInPool) ccerr(ENOTSOCK);
  if (argp == NULL) ccerr(EINVAL);
  if (*argp != 0) sckPool[sck].opt |= cmd; else sckPool[sck].opt &= ~cmd;
err:
  lastSocketError = cc;
  if (debuglevel) printf("soioctl : return <%d>\n",cc);
  return cc ? -1 : 0;
}

/* return socket resources to the pool */
int soclose (int sck)
{
  int i, cc=0;
  /* don't worry about this yet */
  if (sck < 0 || sck >= nSckInPool) ccerr(ENOTSOCK);

  if (sckPool[sck].sckRcvPackets != NULL) SystemFree( (void *) sckPool[sck].sckRcvPackets );
  if (sckPool[sck].sckSndPackets != NULL) SystemFree( (void *) sckPool[sck].sckSndPackets );

  switch( sckPool[sck].typ )
  {
   case SOCK_STREAM:
     tcp_close( sckPool[sck].tcppcb );
     sckPool[sck].tcppcb = NULL;
     break;

   case SOCK_DGRAM:  
     udp_remove( sckPool[sck].udppcb );
     sckPool[sck].udppcb = NULL;
     break;
 
   default:
     ccerr(-1);
     break;
  }

  /* SW! commented out 17.11.03
  if (sckPool[sck].sck != 0)
  {
    nr_plugs_destroy(sckPool[sck].sck);
    sckPool[sck].sck = 0;
  }
  */
err:
  lastSocketError = cc;
  if (debuglevel) printf("soclose : return <%d>\n",cc);
  return cc ? -1 : 0;
}
/*
Parameters
s [in] Descriptor identifying a socket that has been placed in a listening state with the listen function. The connection is actually made with the socket that is returned by accept. 
addr [out] Optional pointer to a buffer that receives the address of the connecting entity, as known to the communications layer. The exact format of the addr parameter is determined by the address family established when the socket was created. 
addrlen [out] Optional pointer to an integer that contains the length of addr. 
*/
int accept(SOCKET sck, struct sockaddr *addr, int *addrlen)
{
  int i, cc=0;
  if (sck < 0 || sck >= nSckInPool) ccerr(ENOTSOCK);
err:
  lastSocketError = cc;
  if (debuglevel) printf("accept : return <%d>\n",cc);
  return cc ? -1 : 0;
}

/*
Parameters
sck [in] Descriptor identifying an unconnected socket. 
name [in] Name of the socket to which the connection should be established. 
namelen [in] Length of name. 
*/
int connect(SOCKET sck,const struct sockaddr *name, int namelen)
{
  int i, cc=0;
  if (sck < 0 || sck >= nSckInPool) ccerr(ENOTSOCK);
err:
  lastSocketError = cc;
  if (debuglevel) printf("connect : return <%d>\n",cc);
  return cc ? -1 : 0;
}
/*
sck [in] Descriptor identifying a bound, unconnected socket. 
backlog [in] Maximum length of the queue of pending connections. If set to SOMAXCONN, the underlying service provider responsible for socket s will set the backlog to a maximum reasonable value. There is no standard provision to obtain the actual backlog value. 
*/
int listen(SOCKET sck, int backlog)
{
  int i, cc=0;
  if (sck < 0 || sck >= nSckInPool) ccerr(ENOTSOCK);
err:
  lastSocketError = cc;
  if (debuglevel) printf("connect : return <%d>\n",cc);
  return cc ? -1 : 0;
}
/*
socket  [in] A descriptor identifying a bound socket. 
buf     [out] A buffer for the incoming data. 
len     [in] The length of buf. 
flags   [in] An indicator specifying the way in which the call is made. 
from    [out] An optional pointer to a buffer that will hold the source address upon return. 
fromlen [in/out] An optional pointer to the size of the from buffer. 
*/
int recvfrom(int sck, char *buf, int len, int flags, struct sockaddr *from, int *fromlen)
{
  int sckPtr, plen, cc=0;
  int bytesread=0;
  if (sck < 0 || sck >= nSckInPool) ccerr(ENOTSOCK);

  if (!sckPool[sck].initialized)
  {
    /* if it's not initialized yet, then I haven't bound it ! */
    if ((cc=bind(sck,NULL,0)) != 0) ccerr(cc);
  }
  if (sckPool[sck].nrcvready > 0)
  {
    sckPool[sck].nrcvready--;
    sckPtr = sckPool[sck].rcvBufferTail;
    plen = MIN(len,sckPool[sck].sckRcvPackets[sckPtr].bufMsgSize);
    if (debuglevel > 1)
      printf("sck %d ptr %d recvfrom %d bytes (min of %d and %d)\n",
              sck,sckPtr,plen,len,sckPool[sck].sckRcvPackets[sckPtr].bufMsgSize);
    memcpy(buf,sckPool[sck].sckRcvPackets[sckPtr].buf,plen);
    if (fromlen != (int *)NULL && (unsigned)(*fromlen) < sizeof(struct sockaddr_in)) ccerr(ENOBUFS);	//271003, grev
    if (from != NULL) *from = *(struct sockaddr *)&sckPool[sck].sckRcvPackets[sckPtr].sckAddr;
    sckPool[sck].rcvBufferTail = (sckPool[sck].rcvBufferTail + 1) % sckPool[sck].numRcvBuffers;
    bytesread = plen;
  }
  else
  {
    if (sckPool[sck].opt & FIONBIO) cc = EWOULDBLOCK;
    /* we're only going to support NON-BLOCKING IO  here (at least for now) */
  }
  
err:
  lastSocketError = cc;
  if (debuglevel) printf("recvfrom : bytes %d return <%d>\n",bytesread,cc);
  return cc ? -1 : bytesread;
}
/*
socket  [in] A descriptor identifying a socket. 
buf     [in] A buffer containing the data to be transmitted. 
len     [in] The length of the data in buf. 
flags   [in] An indicator specifying the way in which the call is made. 
to      [in] An optional pointer to the address of the target socket. 
tolen   [in] The size of the address in to. 
*/
int sendto (int sck, char *buf, int len, int flags, struct sockaddr *to, int tolen)
{
  int sckPtr, cc=0;
  int bytessent=0;
  u_long ipto;
  u8_t resultfree;

  // lwip functions 
  struct ip_addr ip;
  struct pbuf *pbuffer;

  // save old values 
  struct ip_addr rem_ip;
  u16_t remote_port;
  struct sockaddr_in *sa = (struct sockaddr_in *)to;
 
  if (sck < 0 || sck >= nSckInPool) ccerr(ENOTSOCK);

  if (!sckPool[sck].initialized)
  {
    /* if it's not initialized yet, then I haven't bound it ! */
    if ((cc=bind(sck,NULL,0)) != 0) ccerr(cc);
  }

  sckPtr = sckPool[sck].sndBufferHead = (sckPool[sck].sndBufferHead + 1) % sckPool[sck].numSndBuffers;
  if (sa == NULL) sa = &sckPool[sck].sckRcvPackets[sckPtr].sckAddr;
  ipto = htonl(makeIP(sa->sin_addr.S_un.S_un_b.s_b1,sa->sin_addr.S_un.S_un_b.s_b2,
                          sa->sin_addr.S_un.S_un_b.s_b3,sa->sin_addr.S_un.S_un_b.s_b4) );

  /*
    future: this should go into a send queue which gets checked regularly.
            the tail should be incremented following a successful send
            and the head shouldn't be incremented unless there's room
    Right now, we're just sending it out as it comes and hoping for the best ...
  */
  if (debuglevel > 0)
  {
    printf("send to %d.%d.%d.%d:%d (%x %d) on socket %d\n",
	  (u_char)sa->sin_addr.S_un.S_un_b.s_b1,
	  (u_char)sa->sin_addr.S_un.S_un_b.s_b2,
	  (u_char)sa->sin_addr.S_un.S_un_b.s_b3,
	  (u_char)sa->sin_addr.S_un.S_un_b.s_b4,
	  sa->sin_port,
	  ipto, 
	  htons(sa->sin_port),
	  sck);
  }
  sckPool[sck].sndBufferTail = sckPool[sck].sndBufferHead;

  memcpy( &rem_ip, &sckPool[sck].udppcb->remote_ip, sizeof(rem_ip) );
  remote_port = sckPool[sck].udppcb->remote_port;
  
  // allocate a pbuf
  pbuffer = pbuf_alloc(PBUF_IP, len, PBUF_RAM);
  if (pbuffer == NULL) ccerr(ENOMEM);
  memcpy( pbuffer->payload, (const void *) buf, len);

  IP4_ADDR( &ip, sa->sin_addr.S_un.S_un_b.s_b1, sa->sin_addr.S_un.S_un_b.s_b2, sa->sin_addr.S_un.S_un_b.s_b3, sa->sin_addr.S_un.S_un_b.s_b4);
  udp_connect( sckPool[sck].udppcb, &ip, sa->sin_port);
  udp_send( sckPool[sck].udppcb, pbuffer);

  // restore old values
  udp_connect( sckPool[sck].udppcb, &rem_ip, remote_port);

  resultfree = 1;
  while( resultfree > 0)
  {
   resultfree = pbuf_free( pbuffer );
   printf("resultfree pbuf %d\n", resultfree);
  }

  /* SW! commented out 17.11.03
  if ((cc=nr_plugs_send_to( 
			    sckPool[sck].sck,
			    buf,
			    len,
			    plugs_debuglevel, 
                ipto,
			    nr_h2n16(sa->sin_port)
			)) == 0) bytessent = len;
  */
err:
  if (cc) perror("send_to");
  if (debuglevel) printf("sendto : bytes %d return <%d>\n",bytessent,cc);
  lastSocketError = cc;
  return cc ? -1 : bytessent;
}

static void tcpip_init_done(void *arg)
{
	OS_EVENT *pSem;

	pSem = (OS_EVENT *)arg;
	OSSemPost(pSem);
}

void doLWIPInit()
{
        INT8U err;
	struct netif *netif;
	struct ip_addr ipaddr, netmask, gw;
        OS_EVENT *pSem;

  /* lwip minimal init */
        sys_init();						/* Initialize lwip */
        mem_init();
        memp_init();
        pbuf_init();
        netif_init();					/* must be done before tcpip_init() */
  /* lwip minimal end */

        eth_interfaces_init(11);	/* Start the ethernet service task (priority 11) */

	printf("TINE NetTask: starting TCP/IP\n");
	pSem = OSSemCreate(1);		/* create tcpip_init_done semaphore */
	tcpip_init(tcpip_init_done, pSem);
	OSSemPend(pSem, 0, &err);
	OSSemDel(pSem, OS_DEL_ALWAYS, &err);
	printf("TCP/IP initialized.\n");

	printf ("MAC Address: %02X-%02X-%02X-%02X-%02X-%02X\n", ETHADDR0, ETHADDR1, ETHADDR2,ETHADDR3, ETHADDR4, ETHADDR5);

	printf ("Setting IP address to: %d.%d.%d.%d\n", IPADDR0, IPADDR1, IPADDR2, IPADDR3);
	printf ("Setting netmask to: %d.%d.%d.%d\n", NETMASK0, NETMASK1, NETMASK2, NETMASK3);
	printf ("Setting gateway address to: %d.%d.%d.%d\n", GWADDR0, GWADDR1, GWADDR2, GWADDR3);

	IP4_ADDR(&ipaddr, IPADDR0, IPADDR1, IPADDR2, IPADDR3);
	IP4_ADDR(&netmask, NETMASK0, NETMASK1, NETMASK2, NETMASK3);
	IP4_ADDR(&gw, GWADDR0, GWADDR1, GWADDR2, GWADDR3);

	netif = netif_add(&ipaddr, &netmask, &gw, &device_table[ETH_0], tcpip_input);
	netif_set_default(netif);

        // SW! 25.11.03 printf("NetTask: ethernetif_init 0x%08x, tcpip_input 0x%08x\n", ethernetif_init, tcpip_input);
        printf("NetTask: ethernetif_init tcpip_input 0x%08x\n", tcpip_input);
}



