#ifndef	SOCKET_H_
#define	SOCKET_H_

typedef	unsigned short u_short;
typedef unsigned long u_long;
typedef unsigned char u_char;

typedef u_short n_short;		/* short as received from the net */
typedef u_long	n_long;			/* long as received from the net */

#define	SOCK_STREAM	1		/* stream socket */
#define	SOCK_DGRAM	2		/* datagram socket */
#define	SOCK_RAW	3		/* raw-protocol interface */

struct sockaddr {
	short	sa_family;		/* address family */
	char	sa_data[14];		/* up to 14 bytes of direct address */
};

#define SOCK_ISCONN 100
#define SOCK_CLOSEDWITHDATA 101
#define SOCK_ISAVAILABLE 102

/****************************************************************************/

#define	SO_DEBUG	0x0001		/* turn on debugging info recording */
#define	SO_ACCEPTCONN	0x0002		/* socket has had listen() */
#define	SO_REUSEADDR	0x0004		/* allow local address reuse */
#define	SO_KEEPALIVE	0x0008		/* keep connections alive */
#define	SO_DONTROUTE	0x0010		/* just use interface addresses */
#define	SO_BROADCAST	0x0020		/* permit sending of broadcast msgs */
#define	SO_USELOOPBACK	0x0040		/* bypass hardware when possible */
#define	SO_LINGER	0x0080		/* linger on close if data present */
#define	SO_OOBINLINE	0x0100		/* leave received OOB data in line */
#define SO_SMALL	0x0200	
#define	SO_LARGE	0x0400
#define	SO_DONTLINGER	0x0800		/* This used to be ~SO_LINGER */

#define SO_SNDBUF	0x1001		/* send buffer size */
#define SO_RCVBUF	0x1002		/* receive buffer size */
#define SO_SNDLOWAT	0x1003		/* send low-water mark */
#define SO_RCVLOWAT	0x1004		/* receive low-water mark */
#define SO_SNDTIMEO	0x1005		/* send timeout */
#define SO_RCVTIMEO	0x1006		/* receive timeout */
#define	SO_ERROR	0x1007		/* get error status and clear */
#define	SO_TYPE		0x1008		/* get socket type */

struct	linger {
	int	l_onoff;		/* option on/off */
	int	l_linger;		/* linger time */
};

#define	ON	1
#define	OFF	0

#define	SOL_SOCKET	0xffff		/* options for socket level */

struct sockproto {
	short	sp_family;		/* protocol family */
	short	sp_protocol;		/* protocol within family */
};

#define	PF_INET		2		/* internetwork: UDP, TCP, etc. */
#define	AF_INET		PF_INET		/* internetwork: UDP, TCP, etc. */

struct timeval {
	long	tv_sec;		/* seconds */
	long	tv_usec;	/* and microseconds */
};

#define	timerisset(tvp)		((tvp)->tv_sec || (tvp)->tv_usec)
#define	timercmp(tvp, uvp, cmp)	\
	((tvp)->tv_sec cmp (uvp)->tv_sec || \
	 (tvp)->tv_sec == (uvp)->tv_sec && (tvp)->tv_usec cmp (uvp)->tv_usec)
#define	timerclear(tvp)		(tvp)->tv_sec = (tvp)->tv_usec = 0

/****************************************************************************/

#define	MSG_OOB		0x1		/* process out-of-band data */
#define	MSG_PEEK	0x2		/* peek at incoming message */
#define	MSG_DONTROUTE	0x4		/* send without using routing tables */

typedef long	fd_mask;

//#define	FD_SETSIZE	128
#define	NBBY		8				/* bits in a byte */
//#define NFDBITS		(long)(sizeof(fd_mask) * NBBY)	/* bits per mask */
#define	howmany(x, y)	(((x)+((y)-1))/(y))

//typedef	struct fd_set {
//  fd_mask	fds_bits[howmany(FD_SETSIZE, NFDBITS)];
//} fd_set;

//#define	FD_SET(n, p)	((p)->fds_bits[(n)/NFDBITS] |= (((unsigned long)1) << ((n) % NFDBITS)))
//#define	FD_CLR(n, p)	((p)->fds_bits[(n)/NFDBITS] &= ~(((unsigned long)1) << ((n) % NFDBITS)))
//#define	FD_ISSET(n, p)	((p)->fds_bits[(n)/NFDBITS] & (((unsigned long)1) << ((n) % NFDBITS)))
//#define FD_ZERO(p)	bzero((char *)(p), sizeof(*(p)))

#define	FIONREAD	26239	/* get # bytes to read */
#define	FIONBIO		26238	/* set/clear non-blocking i/o */

#define	SIOCATMARK	29477	/* at oob mark? */

#define	SIOCURGRTN	3

struct	urg_anr {
	long	urg_handler;
};

struct iovec {
	char  *iov_base;
	unsigned int	iov_len;
};

#define MAXBUF		1024
/* #define BUFSIZ		512 */
#define MAXALIASES	35
#define MAXADDRESSES	5

#define	h_addr	h_addr_list[0]

struct	hostent {
	char	*h_name;			/* official name of host */
	char	**h_aliases;			/* alias list */
	int 	h_addrtype;			/* host address type */
	int 	h_length;			/* length of address */
	char	**h_addr_list;			/* list of addresses */
	u_long   h_addr_zero;			/* INTERNAL USE */
	char	*arglist[256];			/* INTERNAL USE */
	char	hbuff[MAXBUF];			/* INTERNAL USE */
	char	*h_host_aliases[MAXALIASES];	/* INTERNAL USE */
	char	*h_addrlist[MAXADDRESSES];	/* INTERNAL USE */
	u_long	h__addrlist[MAXADDRESSES];	/* INTERNAL USE */
};

struct	servent {
	char	*s_name;			/* official service name */
	char	**s_aliases;			/* alias list */
	int	s_port;				/* port # */
	char	*s_proto;			/* protocol to use */
	int	s_length;			/* INTERNAL USE */
	char	sbuff[MAXBUF];			/* INTERNAL USE */
	char	*arglist[256];			/* INTERNAL USE */
	char	*s_serv_aliases[MAXALIASES];	/* INTERNAL USE */
};

struct	netent {
	char	*n_name;			/* official name of net */
	char	**n_aliases;			/* alias list */
	int 	n_addrtype;			/* net address type */
	u_long	n_net;				/* network # */
	int	n_length;			/* INTERNAL USE */
	char	nbuff[MAXBUF];			/* INTERNAL USE */
	char	*arglist[256];			/* INTERNAL USE */
	char	*n_net_aliases[MAXALIASES];	/* INTERNAL USE */
};

struct	protoent {
	char	*p_name;			/* official protocol name */
	char	**p_aliases;			/* alias list */
	int	p_proto;			/* protocol # */
	int	p_length;			/* INTERNAL USE */
	char	pbuff[MAXBUF];			/* INTERNAL USE */
	char	*arglist[256];			/* INTERNAL USE */
	char	*p_proto_aliases[MAXALIASES];	/* INTERNAL USE */
};

#define HOSTFILE_ONLY	1
#define DNS_ONLY	2
#define HOSTFILE_FIRST	3
#define DNS_FIRST	4


extern  int h_errno;	
extern  int resolv_priority;

#define	HOST_NOT_FOUND	1 /* Authoritive Answer Host not found */
#define	TRY_AGAIN	2 /* Non-Authoritive Host not found, or SERVERFAIL */
#define	NO_RECOVERY	3 /* Non recoverable errors, FORMERR, REFUSED, NOTIMP */
#define NO_ADDRESS	4 /* Valid host name, no address, look for MX record */

extern  int dns_errno;

#define FILE_NOT_FOUND 1
#define NS_NOT_RESPONDING 2


/****************************************************************************/

#define NAMESERVER_PORT	53

#define PACKETSZ	512        /* maximum packet size */
#define MAXDNAME	256        /* maximum domain name */
#define MAXCDNAME	255        /* maximum compressed domain name */
#define MAXLABEL	63         /* maximum length of domain label */
#define RRFIXEDSZ	10
#define QFIXEDSZ	4

#define QUERY       0        /* standard query */
#define IQUERY      1        /* inverse query */
#define CQUERYM     2        /* completion query (multiple) */
#define CQUERYU     3        /* completion query (unique) */
#define UPDATEA     100      /* add resource record */
#define UPDATED     101      /* delete resource record */
#define UPDATEM     102      /* modify resource record */
#define ZONEINIT    103      /* initial zone transfer */
#define ZONEREF     104      /* incremental zone referesh */

#define NOERROR     0        /* no error */
#define FORMERR     1        /* format error */
#define SERVFAIL    2        /* server failure */
#define NXDOMAIN    3        /* non existent domain */
#define NOTIMP      4        /* not implemented */
#define REFUSED     5        /* query refused */
#define NOCHANGE    100      /* update failed to change db */

#define T_A         1        /* host address */
#define T_NS        2        /* authoritative server */
#define T_MD        3        /* mail destination */
#define T_MF        4        /* mail forwarder */
#define T_CNAME     5        /* connonical name */
#define T_SOA       6        /* start of authority zone */
#define T_MB        7        /* mailbox domain name */
#define T_MG        8        /* mail group member */
#define T_MR        9        /* mail rename name */
#define T_NULL      10       /* null resource record */
#define T_WKS       11       /* well known service */
#define T_PTR       12       /* domain name pointer */
#define T_HINFO     13       /* host information */
#define T_MINFO     14       /* mailbox information */
#define T_MX        15       /* mail routing information */
#define T_UINFO     100      /* user (finger) information */
#define T_UID       101      /* user ID */
#define T_GID       102      /* group ID */
#define T_AXFR      252      /* transfer zone of authority */
#define T_MAILB     253      /* transfer mailbox records */
#define T_MAILA     254      /* transfer mail agent records */
#define T_ANY       255      /* wildcard match */

#define C_IN        1        /* the arpa internet */
#define C_CHAOS     3        /* for chaos net at MIT */
#define C_ANY       255      /* wildcard match */


typedef struct {
    unsigned short    id;  /* query identification number */
    unsigned    rd:1;      /* recursion desired */
    unsigned    tc:1;      /* truncated message */
    unsigned    aa:1;      /* authoritive answer */
    unsigned    opcode:4;  /* purpose of message */
    unsigned    qr:1;      /* response flag */
    unsigned    rcode:4;   /* response code */
    unsigned    unused:2;  /* unused bits */
    unsigned    pr:1;      /* primary server required (non standard) */
    unsigned    ra:1;      /* recursion available */
    unsigned short    qdcount;    /* number of question entries */
    unsigned short    ancount;    /* number of answer entries */
    unsigned short    nscount;    /* number of authority entries */
    unsigned short    arcount;    /* number of resource entries */
} HEADER;

#define INDIR_MASK    0xc0

struct rrec {
    short           r_zone;         /* zone number */
    short           r_class;        /* class number */
    short           r_type;         /* type number */
    unsigned long   r_ttl;          /* time to live */
    int             r_size;         /* size of data area */
    char           *r_data;         /* pointer to data */
};

extern    unsigned short    getshort();
extern    unsigned long     getlong();

/****************************************************************************/

#define  IPPROTO_ICMP       1        /* control message protocol */
#define  IPPROTO_GGP        2        /* gateway^2 (deprecated) */
#define  IPPROTO_TCP        6        /* tcp */
#define  IPPROTO_PUP        12       /* pup */
#define  IPPROTO_UDP        17       /* user datagram protocol */
#define  IPPROTO_RAW        255      /* raw IP packet */
#define  IPPROTO_MAX        256

#define  IPPORT_ECHO        7
#define  IPPORT_DISCARD     9
#define  IPPORT_SYSTAT      11
#define  IPPORT_DAYTIME     13
#define  IPPORT_NETSTAT     15
#define  IPPORT_FTP         21
#define  IPPORT_TELNET      23
#define  IPPORT_SMTP        25
#define  IPPORT_TIMESERVER  37
#define  IPPORT_NAMESERVER  42
#define  IPPORT_WHOIS       43
#define  IPPORT_MTP         57
#define  IPPORT_TFTP        69
#define  IPPORT_RJE         77
#define  IPPORT_FINGER      79
#define  IPPORT_TTYLINK     87
#define  IPPORT_SUPDUP      95
#define  IPPORT_EXECSERVER  512
#define  IPPORT_LOGINSERVER 513
#define  IPPORT_CMDSERVER   514
#define  IPPORT_BIFFUDP     512
#define  IPPORT_WHOSERVER   513
#define  IPPORT_RESERVED    1024
#define  IPPORT_USERRESERVED 5000

struct in_addr {
    union {
        struct { char s_b1,s_b2,s_b3,s_b4; } S_un_b;
        struct { unsigned short s_w1,s_w2; } S_un_w;
        long S_addr;
    } S_un;
#define    s_addr  S_un.S_addr        /* can be used for most tcp & ip code */
#define    s_host  S_un.S_un_b.s_b2   /* host on imp */
#define    s_net   S_un.S_un_b.s_b1   /* network */
#define    s_imp   S_un.S_un_w.s_w2   /* imp */
#define    s_impno S_un.S_un_b.s_b4   /* imp # */
#define    s_lh    S_un.S_un_b.s_b3   /* logical host */
#define    S_baddr S_un.S_un_b
};

#define INADDR_ANY       (unsigned long)0x00000000
#define INADDR_LOOPBACK  (unsigned long)0x7F000001
#define INADDR_BROADCAST (unsigned long)0xffffffff /* Must be masked */

#define IN_LOOPBACKNET   127                       /* official!      */

struct sockaddr_in {
	short		sin_family;
	unsigned short	sin_port;
	struct in_addr	sin_addr;
	char		sin_zero[8];
};

/****************************************************************************/

#define	IPVERSION	4

#ifdef IP_STRUCTS
struct ip {
	u_char	ip_hl:4,		/* header length */
		ip_v:4;			/* version */
	u_char	ip_tos;			/* type of service */
	short	ip_len;			/* total length */
	u_short	ip_id;			/* identification */
	short	ip_off;			/* fragment offset field */
#define	IP_DF 0x4000			/* dont fragment flag */
#define	IP_MF 0x2000			/* more fragments flag */
	u_char	ip_ttl;			/* time to live */
	u_char	ip_p;			/* protocol */
	u_short	ip_sum;			/* checksum */
	struct	in_addr ip_src,ip_dst;	/* source and dest address */
};
#endif

#define	IPOPT_COPIED(o)		((o)&0x80)
#define	IPOPT_CLASS(o)		((o)&0x60)
#define	IPOPT_NUMBER(o)		((o)&0x1f)

#define	IPOPT_CONTROL		0x00
#define	IPOPT_RESERVED1		0x20
#define	IPOPT_DEBMEAS		0x40
#define	IPOPT_RESERVED2		0x60

#define	IPOPT_EOL		0		/* end of option list */
#define	IPOPT_NOP		1		/* no operation */

#define	IPOPT_RR		7		/* record packet route */
#define	IPOPT_TS		68		/* timestamp */
#define	IPOPT_SECURITY		130		/* provide s,c,h,tcc */
#define	IPOPT_LSRR		131		/* loose source route */
#define	IPOPT_SATID		136		/* satnet id */
#define	IPOPT_SSRR		137		/* strict source route */

#define	IPOPT_OPTVAL		0		/* option ID */
#define	IPOPT_OLEN		1		/* option length */
#define IPOPT_OFFSET		2		/* offset within option */
#define	IPOPT_MINOFF		4		/* min value of above */

#ifdef IP_STRUCTS
struct	ip_timestamp {
	u_char	ipt_code;		/* IPOPT_TS */
	u_char	ipt_len;		/* size of structure (variable) */
	u_char	ipt_ptr;		/* index of current entry */
	u_char	ipt_flg:4,		/* flags, see below */
		ipt_oflw:4;		/* overflow counter */
	union {
		n_long	ipt_time[1];
		struct	ipt_ta {
			struct in_addr ipt_addr;
			n_long ipt_time;
		} ipt_ta[1];
	} ip_tunion;
};
#endif

#define	MAXTTL		255		/* maximum time to live (seconds) */
#define	IPFRAGTTL	15		/* time to live for frag chains */
#define	IPTTLDEC	1		/* subtracted when forwarding */

#define	IP_MSS		576		/* default maximum segment size */

/****************************************************************************/

#ifdef IP_STRUCTS
struct icmp {
	u_char	icmp_type;		/* type of message, see below */
	u_char	icmp_code;		/* type sub code */
	u_short	icmp_cksum;		/* ones complement cksum of struct */
	union {
		u_char ih_pptr;			/* ICMP_PARAMPROB */
		struct in_addr ih_gwaddr;	/* ICMP_REDIRECT */
		struct ih_idseq {
			n_short	icd_id;
			n_short	icd_seq;
		} ih_idseq;
		int ih_void;
	} icmp_hun;
#define	icmp_pptr	icmp_hun.ih_pptr
#define	icmp_gwaddr	icmp_hun.ih_gwaddr
#define	icmp_id		icmp_hun.ih_idseq.icd_id
#define	icmp_seq	icmp_hun.ih_idseq.icd_seq
#define	icmp_void	icmp_hun.ih_void
	union {
		struct id_ip  {
			struct ip idi_ip;
		} id_ip;
		u_long	id_mask;
		char	id_data[1];
	} icmp_dun;
#define	icmp_otime	icmp_dun.id_ts.its_otime
#define	icmp_rtime	icmp_dun.id_ts.its_rtime
#define	icmp_ttime	icmp_dun.id_ts.its_ttime
#define	icmp_ip		icmp_dun.id_ip.idi_ip
#define	icmp_mask	icmp_dun.id_mask
#define	icmp_data	icmp_dun.id_data
};
#endif

#define	ICMP_MINLEN	8				/* abs minimum */
#define	ICMP_MASKLEN	12				/* address mask */
#define	ICMP_ADVLENMIN	(8 + sizeof (struct ip) + 8)	/* min */
#define	ICMP_ADVLEN(p)	(8 + ((p)->icmp_ip.ip_hl << 2) + 8)

#define	ICMP_ECHOREPLY		0		/* echo reply */
#define	ICMP_UNREACH		3		/* dest unreachable, codes: */
#define		ICMP_UNREACH_NET	0		/* bad net */
#define		ICMP_UNREACH_HOST	1		/* bad host */
#define		ICMP_UNREACH_PROTOCOL	2		/* bad protocol */
#define		ICMP_UNREACH_PORT	3		/* bad port */
#define		ICMP_UNREACH_NEEDFRAG	4		/* IP_DF caused drop */
#define		ICMP_UNREACH_SRCFAIL	5		/* src route failed */
#define	ICMP_SOURCEQUENCH	4		/* packet lost, slow down */
#define	ICMP_REDIRECT		5		/* shorter route, codes: */
#define		ICMP_REDIRECT_NET	0		/* for network */
#define		ICMP_REDIRECT_HOST	1		/* for host */
#define		ICMP_REDIRECT_TOSNET	2		/* for tos and net */
#define		ICMP_REDIRECT_TOSHOST	3		/* for tos and host */
#define	ICMP_ECHO		8		/* echo service */
#define	ICMP_TIMXCEED		11		/* time exceeded, code: */
#define		ICMP_TIMXCEED_INTRANS	0		/* ttl==0 in transit */
#define		ICMP_TIMXCEED_REASS	1		/* ttl==0 in reass */
#define	ICMP_PARAMPROB		12		/* ip header bad */
#define	ICMP_TSTAMP		13		/* timestamp request */
#define	ICMP_TSTAMPREPLY	14		/* timestamp reply */
#define	ICMP_IREQ		15		/* information request */
#define	ICMP_IREQREPLY		16		/* information reply */
#define	ICMP_MASKREQ		17		/* address mask request */
#define	ICMP_MASKREPLY		18		/* address mask reply */
#define	ICMP_MAXTYPE		18

/****************************************************************************/

#define RES_RETRYS 4
#define DWORD unsigned long

#define RES_INIT	0x0001	/* address initialized */
#define RES_DEBUG	0x0002	/* print debug messages */
#define RES_AAONLY	0x0004	/* authoritative answers only */
#define RES_USEVC	0x0008	/* use virtual circuit */
#define RES_PRIMARY	0x0010	/* query primary server only */
#define RES_IGNTC	0x0020	/* ignore trucation errors */
#define RES_RECURSE	0x0040	/* recursion desired */
#define RES_DEFNAMES	0x0080	/* use default domain name */
#define RES_STAYOPEN	0x0100	/* Keep TCP socket open */
#define RES_DNSRCH	0x0200	/* search up local domain tree */
#define RES_PTR		0x0200	/* use pointer for name resolution */
#define RES_IQUERY	0x0400	/* use iquery for name resolution */
#define RES_IN		0x0800	/* query for IN class */
#define RES_CHAOS	0x1000	/* query for CHAOS class */

#define RES_DEFAULT	(RES_RECURSE | RES_DEFNAMES | RES_DNSRCH)

#define ADDRSORT	1      	/* enable the address-sorting option */
#define MAXADDR		10     	/* max # addresses to sort by */
#define	MAXNS		8      	/* max # name servers we'll track */
#define	MAXDNSRCH	3      	/* max # default domain levels to try */
#define	RES_TIMEOUT	1      	/* seconds between retries */
#define	RES_TIMESHORT	1	/* seconds between retries */
#define	RES_TIMELONG	2	/* seconds between retries */
#define	LOCALDOMAINPARTS 2     	/* min levels in name that is "local" */

#define RES_SHORTTIME (DWORD)  1   /* perform timeout with 1 sec timeout. */
#define RES_LONGTIME (DWORD)   2   /* perform timeout with 30 sec timeout. */
#define RES_DEFTIME (DWORD)    3   /* perform timeout with default (1) time */
#define RES_STATUS (DWORD)     4   /* returns options field of resolver */
#define RES_SETSTATE (DWORD)   5   /* set options bits in _res.options */
#define RES_RSTSTATE (DWORD)   6   /* clear options bits in _res.options */
#define RES_NODENAME (DWORD)   7   /* return name of my machine */
#define RES_DOMAINNAME (DWORD) 8   /* return name of my domain */
#define RES_SETTIME (DWORD)    9   /* set timeout */
#define RES_GETTIME (DWORD)    10  /* return timeout value */

#define	nsaddr	nsaddr_list[0]		/* for backward compatibility */

struct state {
	long retrans;	 			/* retransmition time interval */
	long retry;				/* number of times to retransmit */
	long options;				/* option flags - see below. */
	long nscount;				/* number of name servers */
	struct	sockaddr_in nsaddr_list[MAXNS];	/* address of name server */
	unsigned short	id;			/* current packet id */
	char	defdname[MAXDNAME];		/* default domain */
	char   *dnsrch[MAXDNSRCH + 1];		/* components of domain to search */
	int    ascount;				/*number of addresses */
	struct in_addr sort_list[MAXADDR];	/*address sorting list */ 
};
extern struct state _res;
extern char *p_cdname(), *p_rr(), *p_type(), *p_class();

#define	MAX_PATHS	10
#define MAX_PATH_LEN	512

struct path_info {
	char * pathptrs[MAX_PATHS];
	char paths[MAX_PATH_LEN];
	char excelan_path[80];
};

/****************************************************************************/

typedef int SOCKET;
#define RCVBUF_MTU 1512
#define MAX_NUM_SCK 16
#define DEFAULT_RCVBUF_SIZE 16
#define DEFAULT_SNDBUF_SIZE  8
#define USER_PORT_BASE 1024

typedef struct sckPacketStruct
{
  u_char buf[RCVBUF_MTU];
  int  bufMsgSize;
  struct sockaddr_in sckAddr;
  u_long h_addr;
} SckPacket;

typedef struct sckMemPoolStruct
{
  SckPacket *sckRcvPackets;
  SckPacket *sckSndPackets;
  int  rcvBufferHead;
  int  rcvBufferTail;
  int  nrcvready;
  int  naccready;
  int  tcpbacklog;
  int  tcpparent;
  int  sndBufferHead;
  int  sndBufferTail;
  int  nsndready;
  int  numRcvBuffers;
  int  numSndBuffers;
  int  sck;
  int  typ;
  int  opt;
  int  flg;
  int  sts;
  int  initialized;
  struct sockaddr_in name;
} SckMemPool;

typedef struct 
{
	int udp_plug;
	net_16 port_remote;
	net_16 port_local;
	net_32 ip_remote;
	net_32 ip_local;
	int connected;

	int bytes_received;
	int bytes_sent;
} udp_status;

/****************************************************************************/

int accept(SOCKET sck, struct sockaddr *addr, int *addrlen);
int bind(int s,const struct sockaddr *name, int namelen);
int connect(SOCKET sck,const struct sockaddr *name, int namelen);
int getmymacaddr (char *, int *);
int getpeername (int, struct sockaddr *, int *);
int getsockname (int, struct sockaddr *, int *);
int getsockopt (int, int, int, char *, int *);
int setsockopt(int s, int level, int optname, const char *optval, int optlen);
int ioctl(int sck,long cmd, u_long *argp);
int listen (int, int);
int recv (int, char *, int, int);
int recvfrom(int sck, char *buf, int len, int flags, struct sockaddr *from, int *fromlen);
int shutdown (int, int);
int send (int, char *, int, int);
int sendto (int sck, char *buf, int len, int flags, struct sockaddr *to, int tolen);
int socket(int af, int type, int protocol);
int select(int nfds, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, const struct timeval *timeout);
int soclose (int sck);
int soread (int, char *, int);
int sowrite (int, char *, int);

//void bcopy (char *, char *, int);
//void bzero (char *, int);
//int bcmp  (char *, char *, int);

int loaded(void);
int getbootpdata(char *, int);
u_long getmyipaddr(void);
u_long getsubnetmask(void);

int sethostent(int);
int endhostent(void);
int setnetent(int);
int endnetent(void);
int setprotoent(int);
int endprotoent(void);
int setservent(int);
int endservent(void);

unsigned short htons (unsigned short);
unsigned short ntohs (unsigned short);
unsigned long  ntohl (unsigned long);
unsigned long  htonl (unsigned long);

u_long inet_addr (char *);
u_long inet_network (char *);

int Get_Path(char *, char *, int);
int SetResolvPriority(int);

/****************************************************************************/

char *inet_ntoa(struct in_addr in);
struct in_addr inet_makeaddr (unsigned long, unsigned long);
unsigned long inet_netof (struct in_addr);
unsigned long inet_lnaof (struct in_addr);
struct hostent *gethostbyname(const char *hstname);
struct hostent *gethostbyaddr(const char *hstaddr, int hstlen, int hsttype);
struct hostent *gethostent(void);
struct netent *getnetbyname(char *);
struct netent *getnetbyaddr(unsigned long, int);
struct netent *getnetent(void);
struct protoent *getprotobyname(char *);
struct protoent *getprotobynumber(int);
struct protoent *getprotoent(void);
struct servent *getservbyname(char *, char *);
struct servent *getservbyport(int, char *);
struct servent *getservent(void);
int res_init(void);
int res_mkquery(int,char *,int,int,char *,int,struct rrec *, char *,int);
int res_send(char *, int, char *, int);
int dn_expand( char *, char *, char *, char *, int );
int dn_comp (char *, char *, int, char **, char **);
u_long rhost(char **);
char *raddr(long);
int soperror(char *);

#endif /* SOCKET_H_ */

