/* Patch to make newlib's malloc() and free() thread safe with uC/OS-II,
   and to correct the problem of heap allocation from uC/OS tasks.

   Include this file with your application, and call the initialization
	routine: MallocLockInit(priority) from the main() routine (see note below)

   Copyright 2003 Microtronix
*/


#include <_ansi.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "includes.h"
#include <ucosii/os_cpu.h>
#include <stdio.h>

OS_EVENT	*_malloc_mutex;
int    _malloc_locks;

/*
void *malloc( size_t size )
{
 return( ucos_malloc( size ) );
}

void free( void *data )
{
 ucos_free( data );
}
*/


///////////////////////////////////////////////////////////////////////////////////
// High level API END
///////////////////////////////////////////////////////////////////////////////////


/* This should be called from main() after OSInit().  TaskPriority MUST be a
   priority (not already in use) that is higher than any task calling malloc() 
	or free().
*/
void MallocLockInit(INT8U prio)
{
	INT8U	err;
	_malloc_mutex=OSMutexCreate(prio,&err);
#if 0
	if (err != OS_NO_ERR)
	  printf("Failed to create malloc lock.  err = %d\n", err);
#endif	
	_malloc_locks=0;
}

#define MALLOC_DIRECT


#ifdef MALLOC_DIRECT
/* These routines are replacements for the null versions in newlib.  They are
   called by malloc() and free() to prevent pre-emption corruption of global
   data.
*/
void
__malloc_lock (ptr)
struct _reent *ptr;
{
	INT8U	err;
	int task_prio, mowner_prio, mutex_pip;

    if (!_malloc_mutex) {
#if 0
  	  printf("Failed to lock malloc. err = %d\n", err);
#endif	
      return;
    }
    asm("PFX 8 \n WRCTL %g0;");        /* Disable interrupts                        */

    /* get the task priority, owner priority and pip of the mutex */
    task_prio = OSTCBCur->OSTCBPrio;
    mowner_prio = (INT8U)(_malloc_mutex->OSEventCnt & 0x0FF);
    mutex_pip = (INT8U)(_malloc_mutex->OSEventCnt >> 8);
    
    asm("PFX 9 \n WRCTL %g0;");        /* Enable  interrupts                        */
    
    /* check if we already own this mutex */
    if ((mowner_prio == task_prio) || (mutex_pip == task_prio) ) {
    	/* we do, just count */
    	_malloc_locks++;
    }
    else {
    	/* we don't own it, try to get it*/
    	OSMutexPend(_malloc_mutex,0,&err);
    	if (err != OS_NO_ERR)
#if 0
    	  printf("Failed to lock malloc. err = %d\n", err)
#endif	
          ;
    	else 
          _malloc_locks=1;
   }
}

void
__malloc_unlock (ptr)
struct _reent *ptr;
{
   if( (--_malloc_locks) == 0 ) {
      OSMutexPost(_malloc_mutex);
   }
}

#else

/* As an alternative to the above, these routines are protection wrappers for
   malloc() and free().
*/
void *ucos_malloc(size_t size)
{
	INT8U	err;
	void	*ptr;
	OSMutexPend(_malloc_mutex,0,&err);
	if (err != OS_NO_ERR) {
#if 1
		printf("Failed to lock malloc. err = %d\n", err);
#endif	
		return NULL;
	}
	ptr = malloc(size);
	OSMutexPost(_malloc_mutex);
	return ptr;
}

void ucos_free(void *ptr)
{
	INT8U err;
	OSMutexPend(_malloc_mutex,0,&err);
	if (err != OS_NO_ERR) {
#if 1
		printf("Failed to lock free. err = %d\n", err);
#endif	
		return;
	}
	free(ptr);
	OSMutexPost(_malloc_mutex);
}

#endif


/* This version of _sbrk() replaces the Altera libnios32 version.
   _sbrk() allocates more heap space for malloc without overrunning the
	stack.  In UC/OS, the startup heap (between end-of-program-data and
	top-of-stack) is used by all tasks.  The upper limit of the heap is just
	below the startup stack which will no longer be used by uC/OS tasks.
*/
caddr_t _sbrk(incr)
int incr;
{
  extern char *RAMLimit;
  char *prev_heap_end, *new_heap_end;

  prev_heap_end = RAMLimit;
  new_heap_end = prev_heap_end + incr + 3;
  new_heap_end = (char*)((unsigned int)new_heap_end & (-4));

/* Original: the upper limit of the heap = current stack pointer
  if ((unsigned int)new_heap_end > register(%sp)
    return -1;
*/

/* uC/OS: the upper limit of the heap is the top of the global stack less the
   startup stack area which is no longer used.
*/
  if ((unsigned int)new_heap_end > (unsigned int)nasys_stack_top - 4096)
    return (caddr_t)(-1);
    
  RAMLimit = new_heap_end;
  return (caddr_t)prev_heap_end;
}


///////////////////////////////////////////////////////////////////////////////////
// High level API for TINE protocol by Stefan Weisse / DESY Zeuthen
///////////////////////////////////////////////////////////////////////////////////


void *SystemCalloc(size_t nr, size_t size)
{
 size_t fullsize = nr*size;
 void *ptr;

 printf("calloc %d bytes\n", fullsize);

 ptr = malloc(fullsize);

 printf("calloc res=%x\n", ptr);

 if (ptr != NULL) memset( ptr, 0, fullsize );

 return( ptr );
}

void *SystemMalloc( size_t fullsize )
{
 void *ptr;

 printf("malloc %d bytes\n", fullsize );
 
 ptr = malloc( fullsize );

 printf("malloc res=%x\n", ptr);
 
 return( ptr );
}

void SystemFree( void *ptr )
{
 free( ptr );
}
