#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "srvcore.h"
#include "srvipx.h"
#include "globals.h"
#include "dbtools.h"
#include "ttycmd.h"
#ifdef IPXBSD
#include <netipx/ipx.h>
#include "srvip.h"
#endif

#ifdef IPXNW
IPXECB *CreateIPXECB(int socket);
SPXECB *clnECB[NUM_CLN_ECB];
IPXECB *cnsECB[NUM_CNS_ECB];
IPXECB *mcstECB[NUM_MCST_ECB];

#define ListenForConnection(n) {\
  for (b=0; b<n; b++) if (!(clnECB[b]->ecb.inUseFlag)) {\
    clnECB[b]->hdr.dataStreamType=0;\
    clnECB[b]->ecb.fragmentCount = 2;\
    SPXListenForSequencedPacket(&clnECB[b]->ecb);\
  }\
}
#endif

void closeIPXClients (void);
int initIPXClient (int socket);

extern int clientTimeout;

int ipxClnSck = 0;        /* IPX connections */
int spxClnSck = 0;          /* SPX connections */
int ipxMcastClnSck = 0;   /* IPX connections */

void closeIPXClients (void)
{
  int i;
  char str[256];
  if (!ipxLoaded) return;

# ifdef IPXNW
  if (ipxClnSck) IPXCloseSocket(ipxClnSck);
  if (spxClnSck) IPXCloseSocket(spxClnSck);
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if (FecTbl[conTbl[i]->fecIdx].inetProtocol == IPX)
    {
      sprintf(str,"IPX link to %s %s terminated",
             conTbl[i]->expName, conTbl[i]->prpName);
      feclog(str);
    }
  }
# else
  if (ipxClnSck)
  {
     close(ipxClnSck);
     ipxClnSck = 0;
  }
  if (gListenForMCasts && ipxMcastClnSck)
  {
     close(ipxMcastClnSck);
     ipxMcastClnSck = 0;
  }
# endif
  return;
}

int initIPXClient(int socket)
{
  int i;

  if (!ipxLoaded) return ipx_not_supported;
# ifdef IPXNW
  if (IPXOpenSocket((BYTE *)&socket,0)) return ipx_socket_error; /* open client socket */

  for (i=0; i<NUM_CNS_ECB; i++)           /* create consumer ECBs   */
  {
    if ((cnsECB[i]=CreateIPXECB(PCSocket)) == NULL) return out_of_client_memory;
    IPXListenForPacket(&cnsECB[i]->ecb);
  }

  if (nglobals)                 /* if FEC needs to listen for globals */
  {
    socket = *(int *)&ipxClnSckB[0];
    if (IPXOpenSocket((BYTE *)&socket,0)) return ipx_socket_error;
    for (i=0; i<NUM_GLB_ECB; i++)
    {
      if ((glbECB[i]=CreateIPXECB(socket)) == NULL) return out_of_client_memory;
      IPXListenForPacket(&glbECB[i]->ecb);
    }
  }
  if (gListenForMCasts)           /* if FEC needs to listen for mcasts */
  {
    socket = *(int *)&ipxMcastSckB[0];
    if (IPXOpenSocket((BYTE *)&socket,0)) return ipx_socket_error;
    for (i=0; i<NUM_MCST_ECB; i++)
    {
      if ((mcstECB[i]=CreateIPXECB(socket)) == NULL) return out_of_client_memory;
      IPXListenForPacket(&mcstECB[i]->ecb);
    }
  }
# else
  if (ipxClnSck) close(ipxClnSck);
  ipxClnSck = ipxCreateSocket(0x0, socket);
  if (gListenForMCasts)
  {
    if (ipxMcastClnSck) close(ipxMcastClnSck);
    ipxMcastClnSck = ipxCreateSocket(0x0, *(short *)ipxMcastSckB);
    if (ipxMcastClnSck <= 0) ipxMcastClnSck = 0;
  }
# endif
  return 0;
}
void doIPXConsumer (void)
{
  int i;
  UINT32 prot;
# ifdef IPXBSD
  struct pollfd fds[256];
  int nfds=0, idx, n, maxlen=256;
# endif

  if (!ipxLoaded) return;
  if (tineDebug) sprintf(fecprot, "IPX");

# ifdef IPXNW
  recvIncomingData(NUM_CNS_ECB,cnsECB,doProcessData,IPX);
  if (nglobals) recvIncomingData(NUM_GLB_ECB,glbECB,getGlobalUpdate);
  if (gListenForMCasts) recvIncomingData(NUM_MCST_ECB,mcstECB,doProcessData,IPX);
# else
  ADDSCKTOSET(ipxClnSck);
  ADDSCKTOSET(ipxMcastClnSck);
  FD_ZERO(&ipxconset); FD_SET(ipxClnSck, &ipxconset);
  if (ipxMcastClnSck) FD_SET(ipxMcastClnSck, &ipxconset);
  if (poll(fds,nfds,0) != -1)
  {
     recvIncomingIPXData(ipxClnSck); n = 0;
     if (ipxMcastClnSck) recvIncomingIPXData(ipxMcastClnSck);
  }
  else
  {
    if (errno != EWOULDBLOCK) if (tineDebug > 2)  rprtSckErr("doIPXConsumer","IPX select",ipxClnSck,0,NULL);
  }
  ReturnToClientCycle = 0;
# endif
  return;
}
