#include "project.def"
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include "tine.h"

extern int nofeclog;
extern int ListenForGlobals;
extern int ListenForMCasts;
extern int defaultClientTimeout;
extern int useBracketsInOutput;
extern char outputSeparator[16];
extern int displayErrorMessages;
int tgetEx(char *dev,char *prp,int fmt,int siz,char *tag,int tmr);
int ReportError(int cc);

char *usageTxt[] = {
  "\nSynchronous call: gets the specified TINE device server property\n",
  "tget <device name> <property> [/s=<size> /f=<format> /t=<timeout> /b=<TRUE/FALSE> /c=<separator character> /e=<TRUE/FALSE> /d=<debug level>]",
  "use option /b to surround complex data types with brackets (default TRUE)",
  "use option /c to provide a separator character for array output (default = <single space>)",
  "use option /e to display error messages when the call is not successful (default = TRUE)",
  "e.g. tget /HERA/HEPBLM/WL105K1 AVELOSS",
  "or   tget /HERA/HEPBLM/WL105K1 AVELOSS /s=300 /f=float",
  "\nUse quotes \"\" around device names with spaces",
  "e.g. tget \"/HERA/HEPBPM/WL197 MX\" ORBIT.X",
  "\nIf you regularly call 'tget' inside a script,\nplease make use of a 'trepeater' daemon!",
  "\nA timeout other than the default 1000 msec can be specified via the /t option switch\n",
  "\nSee also: 'tmonitor', 'tsend' for sending and receiving data",
};

#define nUsageLines (int)(sizeof(usageTxt)/sizeof(char *))

int main(int argc, char *argv[])
{
  int i, legacyInput=FALSE;
  int siz = 0, fmt = CF_DEFAULT, tmr=1000;
  char *c, *ptag=NULL, tag[TAG_NAME_SIZE], scratch[256];

  if (argc < 3) 
  {
    for (i=0; i<nUsageLines; i++)
    {
      printf(i == 1 ? "Usage : " : "        ");
      printf("%s\n",usageTxt[i]);
    }
    printf("\n");
    exit(1);
  }
  for (i=3; i<argc; i++)
  { /* options given */
    c = argv[i];
    if (*c++ != '/') 
    {
      legacyInput=TRUE;
      break;
    }
    switch (tolower(*c))
    {
      case 'f':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          fmt = ftoi(scratch);
          if (fmt == CF_STRUCT) if ((ptag=strrchr(scratch,'.')) != NULL) ptag++;
          if (ptag != NULL)
          {
            strncpy(tag,ptag,TAG_NAME_SIZE);
            ptag = tag;
          }
        }
        continue;
      case 's':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          siz = atoi(scratch);      
        }
        continue;
      case 'd':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          tineDebug = atoi(scratch);      
          foregroundTTY = TRUE;
        }
        continue;
      case 't':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          tmr = atoi(scratch); 
        }
        continue;
      case 'b':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          useBracketsInOutput = strtobool(scratch);
        }
        continue;
      case 'c':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          if (!strcmp(c,"\\n")) { c[0] = '\n'; c[1] = 0; }
          if (!strcmp(c,"\\t")) { c[0] = '\t'; c[1] = 0; }
          strncpy(outputSeparator,c,16);
        }
        continue;
      case 'e':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch, c, 32);
          displayErrorMessages = strtobool(scratch);
        }
        continue;
      default:
        continue;
    }
  }
  if (legacyInput)
  {
    if (argc == 5)
    {
      siz = atoi(argv[3]);
      fmt = ftoi(argv[4]);
    }
  }
  gListenForGlobals = gListenForMCasts = FALSE;
  MaxPollingRate = MinPollingRate = 50;
  defaultClientTimeout = 50;
  nofeclog = TRUE;
  SetUserType("CMDLINE");
  return ReportError(tgetEx(argv[1], argv[2], fmt, siz, ptag, tmr));
}
