#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include "tine.h"

typedef struct
{
  int lid;
  char dev[64];
  BYTE *data;
  int fmt;
  int siz;
} MONITOR;

MONITOR *globals = NULL;

void cbMonitor(int id, int cc)
{
  if (cc != 0)
  {
    dbglog("monitor %s : %d",globals[id].dev,cc);
    return;
  }
  switch (globals[id].fmt)
  {
    case CF_DOUBLE:
      dbglog("monitor %s : %g",globals[id].dev,*(double *)globals[id].data);
      break;
    case CF_FLOAT:
      dbglog("monitor %s : %g",globals[id].dev,*(float *)globals[id].data);
      break;
    case CF_LONG:
      dbglog("monitor %s : %d",globals[id].dev,*(SINT32 *)globals[id].data);
      break;
    case CF_SHORT:
      dbglog("monitor %s : %d",globals[id].dev,*(short *)globals[id].data);
      break;
    case CF_TEXT:
    case CF_NAME64:
    case CF_NAME32:
    case CF_NAME16:
      dbglog("monitor %s : %s",globals[id].dev,(char *)globals[id].data);
      break;
    default:
      dbglog("monitor %s : unexpected format %d",globals[id].dev,globals[id].fmt);
      break;
  }
}
int getGlobalsList(char *context,NAME64DBLDBL *list,int *listSize)
{
  DTYPE dout;
  char dev[128];
  int cc;

  memset(&dout,0,sizeof(DTYPE));
  dout.dFormat = CF_NAME64DBLDBL;
  dout.dArrayLength = *listSize;
  dout.data.vptr = list;

  sprintf(dev,"/%s/GLOBALS",context);
  cc = ExecLink(dev,"SRVGLOBALS",&dout,NULL,CA_READ|CA_MUTABLE);
  *listSize = dout.dArrayLength;

  return cc;
}

char *usageTxt[] = {
  "\nMonitors the global parameters from the designated globals server\n",
  "tglobals <context> [/d=<debug level>] or",
  "tglobals [/d=<debug level>] (default context = 'SITE')",
  "e.g. tglobals PETRA",
};
#define nUsageLines (int)(sizeof(usageTxt)/sizeof(char *))
int main(int argc, char *argv[])
{
  char *c, scratch[256];
  char ctx[32],dev[132];
  NAME64DBLDBL ndd[100];
  int i, cc, siz=100;
  DTYPE dout;

  memset(ctx,0,32);
  if (argc == 1) strcpy(ctx,"SITE");
  foregroundTTY = TRUE;
  SetUserType("CMDLINE");
  for (i=1; i<argc; i++)
  { /* options given */
    c = argv[i];
    if (i == 1 && *c++ != '/') 
    { /* legacy input */
      strncpy(ctx,argv[1],32);
      continue;
    }
    switch (tolower(*c))
    {
      case 'd':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          tineDebug = atoi(scratch);      
          foregroundTTY = TRUE;
        }
        continue;
      case 'c':
        strncpy(ctx,c,32);
        continue;
      case '?':
        for (i=0; i<nUsageLines; i++)
        {
          printf(i == 1 ? "Usage : " : "        ");
          printf("%s\n",usageTxt[i]);
        }
        printf("\n");
        exit(1);
      default:
        continue;
    }
  }  
  if (strlen(ctx) == 0) strcpy(ctx,"SITE"); // in case of simple /d=1 args
  cc = getGlobalsList(ctx,ndd,&siz);

  if (cc != 0)
  {
    printf("cannot get globals list : %.32s\n",cc2str(cc));
    exit(1);
  }
  globals = (MONITOR *)calloc(siz,sizeof(MONITOR));
  memset(&dout,0,sizeof(DTYPE));
  for (i=0; i<siz; i++)
  {
    strncpy(globals[i].dev,ndd[i].name,64);
    globals[i].fmt = (int)ndd[i].d1val;
    globals[i].siz = (int)ndd[i].d2val;
    globals[i].data = (BYTE *)calloc(globals[i].siz,fmtsizeof(globals[i].fmt));
    dout.dFormat = globals[i].fmt;
    dout.dArrayLength = globals[i].siz;
    dout.data.bptr = globals[i].data;
    sprintf(dev,"/%s/%s",ctx,globals[i].dev);
    globals[i].lid = recvNetGlobalEx(dev,&dout,cbMonitor,i);
  }

  /* keep the process alive */
  for (;;)
  {
    SystemCycle(TRUE);
  }
  return 0;
}

