#define APPV_MAJOR 2
#define APPV_MINOR 0
#define APPV_REVISION 0 

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include "tine.h"
#include "listener.h"
#include "repeater.h"

char *usageTxt[] = {
  "\ntine repeater: acquire and re-register and export information",
  "from the specified TINE device server\n",
  "tineRepeater <context> <device server> [/c=<new context> /s=<new server name> /f=<fec name> /p=<port offset> /r=<polling interval> /m=<polling mode> /l=<listener table capacity> /x=TRUE /i=TRUE /t=<idle time> /d=<debug level>]\n",
  "if no new context or server are specified, the repeater server will",
  "have the same context as the target server, and a server name given",
  "by the target server name appended with '.RPT'",
  "e.g.",
  "\ttineRepeater HERA BPM\n\n",
  "will produce a server called BPM.RPT in context HERA\n",
  "\ttineRepeater PETRA undulator /c=PETRA.EXT /s=undulator\n\n",
  "will produce a server called undulator in context PETRA.EXT",
  "a port can be specified with the /p switch (default = 101)",
  "a polling interval can be specified with the /r switch (default = 1000 ms)",
  "(note: the polling interval determines the listener refresh rate for any repeated properties)",
  "the listener table capacity can be adjusted with the /l switch (default = 5000 entries)",
  "a polling mode can be specified with the /m switch, e.g. /m=DATACHANGE (default = TIMER)",
  "a specific FEC name can be given with the /f switch (default = Rpt<context%2><server%7>.<port>)",
  "exclusive read can be specified with /x=TRUE",
  "the listening idle time (default = 300 seconds) can be set with the /t switch",
  "  (e.g. /t=86400 will continue listening for 1 day even when the target value is not being read)",
  "ignore listening errors (e.g. via history configuration) can be specified with /i=TRUE"
};

#define nUsageLines sizeof(usageTxt)/sizeof(char *)
int rpPreSystemInit(char *tgtContext,char *tgtServer,char *myContext,char *myServer);
int rpPostSystemInit(char *tgtContext,char *tgtServer,char *myContext,char *myServer);
UINT32 wkAreaSize = 0x100000;
void dumpUsage(void)
{
  int i;
  for (i=0; i<nUsageLines; i++)
  {
    printf(i == 2 ? "Usage : " : "        ");
    printf("%s\n",usageTxt[i]);
  }
  printf("\n");
  exit(1);
}
int main(int argc, char *argv[])
{
  int i, m, p, cc;
  char *c,scratch[64],cntxt[CONTEXT_NAME_SIZE+4],devsrv[EXPORT_NAME_SIZE+4];
  if (argc > 1 && !stricmp(argv[1],RPTR_LCLCTX)) isLocalRepeater = TRUE;
  if (argc < 3 && !isLocalRepeater) dumpUsage();
  memset(thisFec,0,FEC_NAME_SIZE+4);
  memset(cntxt,0,CONTEXT_NAME_SIZE+4);
  memset(devsrv,0,EXPORT_NAME_SIZE+4);
  SetUserType("REPEATER");
  if (isLocalRepeater)
  {
    strncpy(cntxt,RPTR_LCLCTX,CONTEXT_NAME_SIZE);
    strncpy(devsrv,RPTR_LCLSRV,EXPORT_NAME_SIZE);
    strncpy(thisFec,RPTR_LCLFEC,FEC_NAME_SIZE);
    strncpy(thisContext,RPTR_LCLCTX,CONTEXT_NAME_SIZE);
    if ((p=GetPortOffset(thisFec)) < RPTR_DEFAULT_PORT)
    { /* then I must have gotten a free port -> let's try my favorite! */
      p = RPTR_DEFAULT_PORT;
    }
    rpPortOffset = p;
    gIgnoreCentralServices = TRUE;
    gTineConfigurationCoded = TRUE;
    nofeclog = TRUE;
    StartupDebug = FALSE;
  }
  else
  {
    strncpy(cntxt,argv[1],CONTEXT_NAME_SIZE);
    strncpy(devsrv,argv[2],EXPORT_NAME_SIZE-4);
    strcat(devsrv,".RPT");
  }
  for (i=3; i<argc; i++)
  { /* options given */
    c = argv[i];
    if (*c++ != '/') break;
    switch (tolower(*c))
    {
      case 'c':
        if (isLocalRepeater) continue;
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(cntxt,c,CONTEXT_NAME_SIZE);
        }
        continue;
      case 's':
        if (isLocalRepeater) continue;
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(devsrv,c,EXPORT_NAME_SIZE);
        }
        continue;
      case 'f':
        if (isLocalRepeater) continue;
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(thisFec,c,FEC_NAME_SIZE);
        }
        continue;
      case 'd':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          tineDebug = atoi(scratch);      
        }
        continue;
      case 'x':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          hasExclusiveRead = stricmp(scratch,"TRUE") == 0 ? TRUE : FALSE;
        }
        continue;
      case 'i':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          ignoreListeningErrors = stricmp(scratch,"TRUE") == 0 ? TRUE : FALSE;
        }
        continue;
      case 'n':
        if ((c=strchr(c,'=')) != NULL)
        { /* truncate the number of device names to input value */
          c++;
          strncpy(scratch,c,32);
          gTruncateDeviceNames = atoi(scratch);      
        }
        continue;
      case 'r':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          rpPollingRate = atoi(scratch);
          if (rpPollingRate < 100) rpPollingRate = 100;
        }
        continue;
      case 'p':
        if (isLocalRepeater) continue;
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          rpPortOffset = atoi(scratch); 
        }
        continue;
      case 'w':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          wkAreaSize = atoi(scratch); 
        }
        continue;
      case 'l':
        if ((c=strchr(c,'=')) != NULL)
        { /* listener table size */
          c++;
          strncpy(scratch,c,32);
          m = atoi(scratch);
          if (m > 10) MexListenerTableCapacity = m;
        }
        continue;
      case 'm':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          m = GetTransferMode(scratch);
          if (m == CM_NULL) m = CM_TIMER;
          setDefaultListenerMode(m);
          m = BASEMODE(m);
          if (m != CM_TIMER) SetSuppressHeartbeatNotification(FALSE);
          if (m == CM_SINGLE) isSyncRepeater = TRUE;
        }
        continue;
      case 't':
        if ((c = strchr(c, '=')) != NULL)
        { /* idle time */
          int it;
          c++;
          strncpy(scratch, c, 32);
          it = atoi(scratch);
          SetAllowedIdleTime(it);
        }
        continue;
      default:
        continue;
    }
  }
  AssignFecApplicationInfo(APPV_MAJOR,APPV_MINOR,APPV_REVISION);
  if (!isLocalRepeater) printf("Server running (debug level %d)\n",tineDebug);
  SetConnectionTableCapacity(MexListenerTableCapacity);
  SetWorkAreaSize(wkAreaSize);
  FeclogDepth = 500;
# ifdef MULTITHREADED
  useMultiThreadedEquipmentFunctions = FALSE;
  useMultiThreadedBackgroundTasks = FALSE;
# endif
  rpPreSystemInit(argv[1],argv[2],cntxt,devsrv);
  if ((cc=SystemInit(TRUE)) != 0) { printf("init error: %s\n",erlst[cc]); exit(1); }
  rpPostSystemInit(argv[1],argv[2],cntxt,devsrv);
  if (isLocalRepeater) 
  {
    gSystemRunningStandAlone = TRUE;
    sndRegisteredExports(&thisCycleTime);
    gSystemRunningStandAlone = FALSE;
  }
  for (;;)
  {
    SystemCycle(TRUE);
  }
  return 0;
}
