#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include "tine.h"

extern int nofeclog;
extern int ListenForGlobals;
extern int ListenForMCasts;
extern int useBracketsInOutput;
extern char outputSeparator[16];
int tmonitor(char *dev,char *prp);
int tmonitorEx(char *dev,char *prp,int fmt,int siz,char *tag,int tmr);

char *usageTxt[] = {
  "\nAsynchronous call : Monitors the specified TINE device server property\n",
  "tmonitor <device name> <property> [/s=<size> /f=<format> /t=<timer> /b=<TRUE/FALSE> /c=<separator character> /d=<debug level>]",
  "use option /b to surround complex data types with brackets (default TRUE)",
  "use option /c to provide a separator character for array output (default = <single space>)",
  "e.g. tmonitor /HERA/HEPBLM/WL105K1 AVELOSS",
  "or   tmonitor /HERA/HEPBLM/WL105K1 AVELOSS /s=300 /f=float",
  "\nUse quotes \"\" around device names with spaces",
  "e.g. tmonitor \"/HERA/HEPBPM/WL197 MX\" ORBIT.X",
  "\nA polling interval other than the default 1000 msec can be specified via the /t option switch\n",
  "\nSee also: 'tget', 'tsend' for sending and receiving data",
};

#define nUsageLines (int)(sizeof(usageTxt)/sizeof(char *))

int main(int argc, char *argv[])
{
  int i, cc, legacyInput=FALSE;
  int siz = 0, fmt = CF_NULL, tmr=1000;
  char *c, *ptag=NULL, tag[TAG_NAME_SIZE], scratch[256];
  if (argc < 3) 
  {
    for (i=0; i<nUsageLines; i++)
    {
      printf(i == 1 ? "Usage : " : "        ");
      printf("%s\n",usageTxt[i]);
    }
    printf("\n");
    exit(1);
  }
  for (i=3; i<argc; i++)
  { /* options given */
    c = argv[i];
    if (*c++ != '/') 
    {
      legacyInput=TRUE;
      break;
    }
    switch (tolower(*c))
    {
      case 'f':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          fmt = ftoi(scratch);      
          if (fmt == CF_STRUCT) if ((ptag=strrchr(scratch,'.')) != NULL) ptag++;
          if (ptag != NULL)
          {
            strncpy(tag,ptag,TAG_NAME_SIZE);
            ptag = tag;
          }
        }
        continue;
      case 's':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          siz = atoi(scratch);      
        }
        continue;
      case 'd':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          tineDebug = atoi(scratch);      
          foregroundTTY = TRUE;
        }
        continue;
      case 't':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          tmr = atoi(scratch); 
        }
        continue;
      case 'b':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          useBracketsInOutput = (!stricmp(scratch,"TRUE") || !stricmp(scratch,"YES"));
        }
        continue;
      case 'c':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          if (!strcmp(c,"\\n")) { c[0] = '\n'; c[1] = 0; }
          if (!strcmp(c,"\\t")) { c[0] = '\t'; c[1] = 0; }
          strncpy(outputSeparator,c,16);
        }
        continue;
      default:
        continue;
    }
  }
  if (legacyInput)
  {
    if (argc == 5)
    {
      siz = atoi(argv[3]);
      fmt = ftoi(argv[4]);
    }
  }
  MaxPollingRate = MinPollingRate = 200;
  nofeclog = TRUE;
  SetUserType("CMDLINE");
  if ((cc=tmonitorEx(argv[1],argv[2],fmt,siz,ptag,tmr)) != 0) printf("error : %s\n",erlst[cc&0xfff]);
  for (;;) SystemCycle(TRUE);
  return cc;
}
