#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include "tine.h"

extern int nofeclog;
extern int ListenForGlobals;
extern int ListenForMCasts;
extern int defaultClientTimeout;
extern int displayErrorMessages;
int tsendrecv(char *dev,char *prp,char *datastr,int fmt,int siz,int tmr,int acc);
void dumpReturnedSyncData(int length, char *tag);
int ReportError(int cc);
char *usageTxt[] = {
  "\nSynchronous call: sends the input data to the specified TINE device",
  "server and property, and acquires the default output. Sets the WRITE access bit by default.\n",
  "A single string can be sent by enclosing the data in quotes \"\".",
  "A data array can be separated by blanks, commas, or semi-colons.",
  "A string data array will be sent as either NAME32 or NAME16 elements.",
  "Do not enclose string arrays with quotes!\n",
  "You can specify a format by using the /f option switch and you can",
  "specify a specific access (default= READ|WRITE) by using the /a option switch and you can",
  "specify an explicit size by using the /s option switch (cannot exceed the number of elements!)\n",
  "A timeout other than the default 1000 msec can be specified via the /t option switch\n",
  "Usage : tsendrecv <device name> <property> [/f=format /s=size /t=timeout /a=access] <data1> [<data2> <data3> ...]\n",
  "        e.g. tsendrecv /HERA/HEPBLM/WL105K1 MODE 1",
  "        or   tsendrecv /HERA/HEPBPM/WL105K1 MESSAGE \"hello\"",
  "\nUse quotes \"\" around device names with spaces\n",
  "        e.g. tsendrecv \"/HERA/HEPBPM/WL197 MX\" MODE 1",
  "        or   tsendrecv \"/HERA/HEPBPM/WL197 MX\" MESSAGE /a=READ \"hello\"",
  "\nUse option /e=TRUE or FALSE to display error messages when the call is not successfull (default=TRUE)",
  "Use option /d=<debug level> to display verbose debugging information",
  "\nSee also: 'tmonitor', 'tget' for receiving data, 'tsend' for sending data.",
};

#define nUsageLines (int)(sizeof(usageTxt)/sizeof(char *))
int isNumPrefix(char c)
{
  if (c == '.') return TRUE;
  if (c == '+') return TRUE;
  if (c == '-') return TRUE;
  return FALSE;
}
int main(int argc, char *argv[])
{
  int i, len, dargc, fmt=CF_NULL, siz = 0, tmr=1000, acc=CA_READ|CA_WRITE;
  char *s, datastr[4096];
  char *c, scratch[256], **dargv;
  int isText = FALSE;

  MaxNumConnections = 10;
  gListenForGlobals = gListenForMCasts = FALSE;
  MaxPollingRate = MinPollingRate = 50;
  defaultClientTimeout = 50;
  nofeclog = TRUE;
  
  if (argc < 4) 
  {
    for (i=0; i<nUsageLines; i++)
    {
      printf("%s\n",usageTxt[i]);
    }
    printf("\n");
    exit(1);
  }
  dargv = (char **)malloc((argc-3)*sizeof(char *));
  dargc = 0;
  for (i=3; i<argc; i++)
  { /* check for options and data */
    c = argv[i];
    if (*c++ != '/') 
    { /* a data input entry */
      dargv[dargc++] = argv[i];
      continue;
    }
    switch (tolower(*c))
    {
      case 'f':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          fmt = ftoi(scratch);      
        }
        continue;
      case 's':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          siz = atoi(scratch);      
        }
        continue;
      case 'd':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          tineDebug = atoi(scratch);
          foregroundTTY = TRUE;
        }
        continue;
      case 't':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          tmr = atoi(scratch); 
        }
        continue;
      case 'a':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,32);
          acc = getAccessType(scratch);
        }
        continue;
      case 'e':
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch, c, 32);
          displayErrorMessages = strtobool(scratch);
        }
        continue;
      default:
        continue;
    }
  }
  if (dargc == 1)
  {
    s = dargv[0];
    len = (int)strlen(s);
    for (i=0; i<len; i++) if (!isdigit(s[i]) && !isNumPrefix(s[i])) break;
    if (fmt == CF_NULL && (i < len) && strncmp(s,"0x",2) && strncmp(s,"0X",2))
    {
      isText = TRUE;
    }
  }
  datastr[0] = 0;
  if (isText) strcat(datastr,"\"");
  for (i=0; i<dargc; i++) 
  {
    strcat(datastr,dargv[i]);
    if (i < dargc - 1) strcat(datastr," ");
  }
  if (isText) strcat(datastr,"\"");
  SetUserType("CMDLINE");
  return ReportError(tsendrecv(argv[1], argv[2], datastr, fmt, siz, tmr, acc));
}
